<?php

namespace FiloBlu\Refilo\Remote\Connector;

use Elasticsearch\Client;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\DataObjectFactory;

/**
 * Class ElasticSearchConnector
 * @package FiloBlu\Refilo\Remote\Connector
 */
class BumbleSearch extends AbstractConnector
{

    /** @var int */
    protected $pageSize = 1000;

    /**
     * @var Client
     */
    protected $handler;

    /**
     * @var string
     */
    protected $username;

    /**
     * @var string
     */
    protected $password;

    /**
     * @var \Magento\Customer\Model\ResourceModel\Group\Collection
     */
    protected $_customerGroupColl;

    //TODO add attr.title etc and type number
    /**
     * @var array
     */
    protected $base_mapping = [
        [
            'path'     => 'weight',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'updated_at',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => '*.key',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => '*.swatch.*',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'assets.*',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'crosssells',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'meta_*',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'rating.*',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'related',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'upsells',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'variants.*',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'bundle.*',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'custom_options.*',
            'analyzer' => 'disabled'
        ],
        [
            'path'     => 'pricing.*',
            'analyzer' => 'number'
        ],
        [
            'path'     => 'category_data.*',
            'analyzer' => 'number'
        ]
    ];

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * BumbleSearch constructor.
     * @param DataObjectFactory $dataObjectFactory
     * @param \Psr\Log\LoggerInterface $logger
     * @param \Magento\Customer\Model\ResourceModel\Group\Collection $customerGroupColl
     * @param ScopeConfigInterface $scopeConfig
     */
    public function __construct(
        DataObjectFactory $dataObjectFactory,
        \Psr\Log\LoggerInterface $logger,
        \Magento\Customer\Model\ResourceModel\Group\Collection $customerGroupColl,
        ScopeConfigInterface $scopeConfig
    )
    {
        $this->dataObjectFactory = $dataObjectFactory;
        $this->logger = $logger;
        $this->_customerGroupColl = $customerGroupColl;
        $this->scopeConfig = $scopeConfig;
    }

    /**
     * @return mixed|void
     */
    public function disconnect(): ConnectorInterface
    {
        if (!$this->handler) {
            return $this;
        }

        $this->handler = null;

        return $this;
    }

    /**
     * @param array $entity
     * @param string $collection
     * @return mixed|void
     */
    public function delete(array $entity, string $collection)
    {
    }

    /**
     * @param iterable $entities
     * @param string $collection
     * @return mixed|void
     */
    public function update(iterable $entities, string $collection)
    {
        $this->create($entities, $collection, 'update');
    }

    /**
     * @param iterable $entities
     * @param string $collection
     * @return mixed|void
     */
    public function create(iterable $entities, string $collection, $action = 'insert')
    {
        $page_size = (int)$this->scopeConfig->getValue('sync/data/batch_size');
        $this->pageSize = ($page_size) ? $page_size : $this->pageSize;

        if ($this->metadata && $this->metadata->getDataByKey('name') == 'attributes') {
            return $this->mapFields($collection, $entities);
        }

        if ($action == 'insert') {
            $this->createIndex($collection);
        }

        $client = $this->getHandler();

        $i = 1;

        try {

            foreach ($entities as $entity) {

                $doc = $this->cleanDoc($entity);

                $docs[] = $doc;

                if ($i % $this->pageSize === 0) {
                    $response = $client->request('POST', "indexes/{$collection}/bulk?action={$action}", [
                        'json'    => $docs,
                        'headers' => $this->getHeaders()
                    ]);
                    $docs = [];
                }
                $i++;


            }

            if (!empty($docs)) {
                $response = $client->request('POST', "indexes/{$collection}/bulk?action={$action}", [
                    'json'    => $docs,
                    'headers' => $this->getHeaders()
                ]);
            }

        } catch (\Exception $e) {
            echo $e->getMessage();
        }
    }

    /**
     * @param $index
     * @param $entities
     */
    private function mapFields($index, $entities)
    {

        try {
            $client = $this->getHandler();
            $response = $client->request('DELETE', "indexes/{$index}/mapping/field/reset", [
                'headers' => $this->getHeaders()
            ]);

//            $response = $client->request('GET', 'indexes/'.$index . '/fields', [
//                'headers' => $this->getHeaders()
//            ]);
//
//            $fields = $this->decodeResponse($response);

            $mapping = $this->buildMapping($index, $entities);

            $response = $client->request('POST', "indexes/{$index}/mapping", [
                'headers' => $this->getHeaders(),
                'json'    => [
                    'fields' => array_values($mapping)
                ]
            ]);

        } catch (\Exception $e) {
            echo $e->getMessage();
        }
    }

    /**
     * @return Client
     * @throws \Exception
     */
    protected function getHandler()
    {
        $this->connect();
        return $this->handler;
    }

    /**
     * @return ConnectorInterface
     */
    public function connect(): ConnectorInterface
    {

        if (!$this->handler) {
            $hosts = $this->getHosts();

            if (!count($hosts))
                throw new \Exception('no hosts found');

            $this->handler = new \GuzzleHttp\Client([
                    'base_uri' => $hosts[0]
            ]);
        }
        return $this;
    }

    /**
     * @return array
     */
    protected function getHosts(): array
    {
        $hosts = $this->getConfiguration()->getData(self::CONFIGURATION_HOST_POOL);

        $dsns = [];
        $dsn = '';
        foreach ($hosts as $host) {

            if ($host['username']) {
                $this->username = $host['username'];
                $this->password = $host['password'];
            }

            $dsn .= "{$host['scheme']}://";

            $dsn .= $host['host'];

            if ($host['port']) {
                $dsn .= ":{$host['port']}";
            }
            $dsns[] = $dsn;
            break;
        }

        return $dsns;
    }

    private function getHeaders()
    {
        return [
//            'Authorization' => 'Bearer ' . $this->token,
            'Accept'    => 'application/json',
            'X-APP-ID'  => $this->username,
            'X-API-KEY' => $this->password
        ];
    }

    /**
     * @param $fields
     * @param $entities
     * @return array
     */
    private function buildMapping($fields, $entities)
    {

        $mapping = [];
        $filter_price = false;

        foreach ($entities as $entity) {

            $included_attr = ($entity->getData('isFilterable') || $entity->getData('isSortable') || ($entity->getData('searchable')));

            if (!$included_attr)
                continue;

            if ($entity->getData('code') === 'price') {
                $filter_price = $entity;
                continue;
            }

            $attribute_prefix = ($entity->getData('isFilterable')) ? 'filters.' : 'attributes.';

            $field_template = [
                'name'                  => $attribute_prefix . $entity->getData('code') . '.value',
                'label'                 => $entity->getData('title'),
                'code'                  => $entity->getData('code'),
                'facet'                 => ($entity->getData('isFilterable')) ? true : false,
                'facet_size'            => 100,
                'filterable'            => ($entity->getData('isFilterable')) ? true : false,
                'sortable'              => ($entity->getData('isSortable')) ? true : false,
                'searchable'            => ($entity->getData('searchable')) ? true : false,
                'filterable_type'       => 'values',
                'filterable_range_step' => 0,
                'filterable_priority'   => (int)$entity->getData('position'),
                'sortable_priority'     => (int)$entity->getData('sort_weight'),
                'searchable_priority'   => (int)$entity->getData('search_weight')
            ];

            $mapping[$entity->getData('code')] = $field_template;

        }

        if ($filter_price)
            $mapping = $this->buildPriceMapping($mapping, $filter_price);

        return $mapping;

    }

    /**
     * @param $mapping
     * @param $filter_price
     * @return mixed
     */
    private function buildPriceMapping($mapping, $filter_price)
    {

//      $matches  = preg_grep ('/^pricing_.(.*).final_price/i', $fields);
        $customerGroups = $this->_customerGroupColl->toOptionArray();
        $field_template = [];

        if (count($customerGroups)) {
            foreach ($customerGroups as $item) {

                $attribute = "pricing_{$item['value']}";

                $field_template = [
                    'name'                  => "pricing_{$item['value']}.final_price.numeric",
                    'label'                 => $filter_price->getData('title'),
                    'code'                  => "pricing_{$item['value']}",
                    'facet'                 => true,
                    'facet_size'            => 20,
                    'filterable'            => true,
                    'sortable'              => ($filter_price->getData('isSortable')) ? true : false,
                    'searchable'            => false,
                    'filterable_type'       => 'range',
                    'filterable_range_step' => (int)$this->metadata->getData('filterable_range_step'),
                    'filterable_priority'   => 0,
                    'sortable_priority'     => 0,
                    'searchable_priority'   => 0
                ];

                $mapping[$attribute] = $field_template;

            }

        }

        $mapping['pricing'] = [
            'name'                  => 'pricing.final_price',
            'label'                 => $filter_price->getData('title'),
            'code'                  => 'pricing',
            'facet'                 => true,
            'facet_size'            => 20,
            'filterable'            => true,
            'sortable'              => ($filter_price->getData('isSortable')) ? true : false,
            'searchable'            => false,
            'filterable_type'       => 'range',
            'filterable_range_step' => (int)$this->metadata->getData('filterable_range_step'),
            'filterable_priority'   => 0,
            'sortable_priority'     => 0,
            'searchable_priority'   => 0
        ];

        return $mapping;

    }

//    private function decodeHost($connection_string) {
//
//        $connection_parts = explode('|',$connection_string);
//
//        $auth_parts = (isset($connection_parts[0])) ? explode(":",$connection_parts[0]) : [];
//
//        $connection["host"] = (isset($connection_parts[1])) ? $connection_parts[1] : '';
//        $connection["auth"] = ['json' => [
//            "appId" => $auth_parts[0],
//            "accountId" => isset($auth_parts[1]) ? $auth_parts[1] : '',
//            "secret" => isset($auth_parts[2]) ? $auth_parts[2] : '',
//        ]];
//
//        return $connection;
//
//    }

//    /**
//     * @return array
//     */

    /**
     * @param $index
     * @return $this
     */
    public function createIndex($index)
    {

        $client = $this->getHandler();
        try {
            $response = $client->request('GET', 'indexes', [
                'headers' => $this->getHeaders()
            ]);

            $indices = $this->decodeResponse($response);

            if (array_key_exists($index, $indices)) {
                return $this;
            }

            $response = $client->request('POST', "indexes/{$index}", [
                'headers' => $this->getHeaders(),
                'json'    => [
                    'fields' => $this->base_mapping
                ]
            ]);

        } catch (\Exception $e) {
            throw new \Error($e);
        }

        return $this;
    }

    /**
     * @param $response
     * @return mixed
     */
    private function decodeResponse($response)
    {
        return json_decode($response->getBody());
    }

    /**
     * @param $entity
     */
    public function cleanDoc($entity)
    {
        $doc = $entity->toArray();
        $doc['id'] = (string)$entity->getId();

//        if (isset($doc['filters'])) {
//            foreach ($doc['filters'] as $key => $filter) {
//                $doc['attributes'][$key] = $filter;
//            }
//            unset($doc['filters']);
//            unset($doc['sortables']);
//        }

        return $doc;

    }

    /**
     * @return array|mixed|null
     */
    protected function getPrefix()
    {
        return $this->getConfiguration()->getData('prefix');
    }
}
